@@ -4,10 +4,11 @@ import sys
 import os.path
 import time
 import logging
-import urlparse
 import subprocess
 import functools
 import itertools
+import math
+from subprocess import Popen, PIPE
 
 # print 0
 import requests
@@ -15,34 +16,17 @@ from requests_futures.sessions import FuturesSession
 
 import hgapi
 import gitapi
-import OpenSSL
+# import OpenSSL
 
 logging.getLogger("requests").setLevel(logging.WARNING)
 logger = logging.getLogger('bucket')
 
 import exceptions as exc
-from commodity.pattern import Bunch, MetaBunch
+from commodity.pattern import MetaBunch
 from commodity.thread_ import SimpleThreadPool
 
 import lurl
 
-# import math
-# http://stackoverflow.com/questions/5194057/better-way-to-convert-file-sizes-in-python
-# def convertSize(size):
-#     size_name = ("KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB")
-#     i = int(math.floor(math.log(size, 1024)))
-#     p = math.pow(1024, i)
-#     s = round(size / p, 2)
-#     if (s > 0):
-#         return '%s %s' % (s, size_name[i])
-#     else:
-#         return '0B'
-
-
-def convertSize(bytes):
-    k = float(bytes) / 1000
-    return "{:.0f} KB".format(k)
-
 
 bitbucket_org = 'bitbucket.org'
 BB_API1 = 'https://api.bitbucket.org/1.0/'
@@ -52,6 +36,40 @@ PROTO_SSH = 'ssh'
 PROTO_HTTPS = 'https'
 
 
+def git_cmd(path, *args):
+    """Run a git command in path and return the result. Throws on error."""
+    if not path:
+        path = '.'
+
+    proc = Popen(["git"] + list(args), stdout=PIPE, stderr=PIPE, cwd=path)
+    out, err = [x.decode("utf-8") for x in proc.communicate()]
+
+    if proc.returncode:
+        cmd = "git " + " ".join(args)
+        raise gitapi.gitapi.GitException(
+            "Error running %s:\n\tErr: %s\n\tOut: %s\n\tExit: %s"
+            % (cmd,err,out,proc.returncode), exit_code=proc.returncode)
+
+    return out, err
+
+
+def hg_cmd(path, *args):
+    """Run a git command in path and return the result. Throws on error."""
+    if not path:
+        path = '.'
+
+    proc = Popen(["hg"] + list(args), stdout=PIPE, stderr=PIPE, cwd=path)
+    out, err = [x.decode("utf-8") for x in proc.communicate()]
+
+    if proc.returncode:
+        cmd = "hg " + " ".join(args)
+        raise hgapi.hgapi.HgException(
+            "Error running %s:\n\tErr: %s\n\tOut: %s\n\tExit: %s"
+            % (cmd, err, out, proc.returncode), exit_code=proc.returncode)
+
+    return out, err
+
+
 def bb_api1_uri(path):
     uri = BB_API1 + path
     return uri
@@ -62,16 +80,21 @@ def bb_api2_uri(path):
     return uri
 
 
-def reply2json(rep, ok=None):
+def reply2json(reply, ok=None):
     ok = ok or 200
 
-    if rep.status_code != ok:
-        raise exc.RemoteException.from_reply(rep)
+    if reply.status_code != ok:
+        logger.info(reply.content)
+        raise exc.RemoteException.from_reply(reply)
 
-    if not rep.content:
+    if not reply.content:
         return None
 
-    return rep.json()
+    try:
+        return reply.json()
+    except ValueError as e:
+        logger.critical("{}: {}".format(e, reply.text))
+        raise
 
 
 class Account(object):
@@ -79,10 +102,10 @@ class Account(object):
         self.username = username or "anonymous"
         self.password = password
 
-    @classmethod
-    def from_args(cls, *args):
-        assert len(args) == 2
-        return Account(str.join(':', args))
+    # @classmethod
+    # def from_args(cls, *args):
+    #     assert len(args) == 2
+    #     return Account(str.join(':', args))
 
     def is_anonymous(self):
         return self.username == "anonymous"
@@ -118,7 +141,7 @@ class Session(object):
         return self._http_session
 
     def request(self, method, path, data=None, ok=None):
-        self.logger.debug("request %s %s %s", method, path, data)
+        self.logger.debug("request '%s' '%s' '%s'", method, path, data)
         while 1:
             url = path if path.startswith('http') else bb_api2_uri(path)
             try:
@@ -127,7 +150,7 @@ class Session(object):
                 retval = reply2json(reply, ok=ok)
                 break
 
-            except (exc.ServiceUnavailable, OpenSSL.SSL.ZeroReturnError):
+            except exc.ServiceUnavailable:
                 time.sleep(1)
 
             except requests.exceptions.ConnectionError, e:
@@ -137,8 +160,12 @@ class Session(object):
         return retval
 
     def async_request(self, method, path, data=None, ok=None):
-        def to_json(sess, resp):
-            resp.data = resp.json()
+        def to_json(sess, reply):
+            try:
+                reply.data = reply.json()
+            except ValueError as e:
+                self.logger.critical("{}: {}".format(e, reply.text))
+                raise
 
         self.logger.debug("async request %s %s %s", method, path, data)
         url = path if path.startswith('http') else bb_api2_uri(path)
@@ -167,9 +194,16 @@ class Session(object):
         return "<Session '%s'>" % self.account
 
 
+class SyncStatus:
+    UPDATE = "update"
+    CLONED = "cloned"
+    PULLED = "pulled"
+
+
 class Repo(object):
     def __init__(self, manager, full_name=None, dirname=None):
         assert self.__class__ is not Repo
+        assert isinstance(manager, RepoManager)
         self.manager = manager
         self.full_name = full_name
         self.dirname = dirname
@@ -184,8 +218,9 @@ class Repo(object):
             data = MetaBunch(manager.get_repo_info(full_name))
             repo_cls = {'hg': HgRepo, 'git': GitRepo}[data.scm]
             retval = repo_cls(manager, data['full_name'])
+
         except exc.RemoteException as e:
-            if e.reason == "NOT_FOUND":
+            if e.reason.upper() == "NOT_FOUND":
                 raise exc.NoSuchRepo()
             raise
 
@@ -235,6 +270,10 @@ class Repo(object):
     def slug(self):
         return self.full_name.split('/')[1]
 
+    @property
+    def basename(self):
+        return os.path.split(self.dirname)[1]
+
     @property
     def api_repo(self):
         if self._api_repo is None:
@@ -250,31 +289,24 @@ class Repo(object):
         return retval
 
     def __getitem__(self, key):
-        if key == 'size':
-            return self.size
-
         if key == 'webpage':
-            return self.wegpage
+            return self.webpage
 
         if key == 'access':
             return self.access
 
-        return self.data[key]
+        return self._data[key]
 
     def is_private(self):
-        return self.data.is_private
+        return self._data.is_private
 
     @property
     def access(self):
         return 'private' if self.is_private() else 'public'
 
-    @property
-    def size(self):
-        return convertSize(self.data.size)
-
     @property
     def webpage(self):
-        return self.data.links.html.href
+        return self._data.links.html.href
 
     @property
     def scm(self):
@@ -282,7 +314,7 @@ class Repo(object):
             scms = {hgapi.Repo: 'hg', gitapi.Repo: 'git'}
             return scms[self._api_repo.__class__]
 
-        return self.data.scm
+        return self._data.scm
 
     def _get_origin_by_rest(self, proto):
         return self._get_origin_sid(proto)
@@ -291,8 +323,10 @@ class Repo(object):
         assert proto in [PROTO_HTTPS, PROTO_SSH]
 
         try:
-            origin = self._do_get_origin_by_cmd()
-            url = lurl.change_login(urlparse.urlparse(origin))
+            self.origin = self._do_get_origin_by_cmd()
+            # print("======", self.origin)
+            url = lurl.parse(self.origin)
+            url = lurl.change_login(url)
 
             if url.scheme == proto:
                 return url.geturl()
@@ -303,7 +337,7 @@ class Repo(object):
 
     @classmethod
     def _get_full_name(cls, origin):
-        url = urlparse.urlparse(origin)
+        url = lurl.parse(origin)
         if url.hostname != bitbucket_org:
             return None
 
@@ -314,7 +348,7 @@ class Repo(object):
         return full_name.strip('/')
 
     def _apply_pwd(self, origin):
-        url = urlparse.urlparse(origin)
+        url = lurl.parse(origin)
         if url.scheme == PROTO_HTTPS:
             username = self.manager.session.account.username
             password = self.manager.session.account.password
@@ -331,6 +365,16 @@ class Repo(object):
         except (hgapi.hgapi.HgException, gitapi.gitapi.GitException, UnicodeDecodeError) as e:
             raise exc.CloneError(e)
 
+    def commited(self):
+        return
+
+    def pushed(self):
+        return
+
+    def __cmp__(self, other):
+        assert isinstance(other, Repo)
+        return cmp(self.full_name, other.full_name)
+
     def __eq__(self, other):
         return self.scm == other.scm and self.full_name == other.full_name
 
@@ -375,10 +419,12 @@ class HgRepo(Repo):
         self._api_repo.hg_id()
 
     def _get_origin_sid(self, proto=PROTO_HTTPS):
+        username = self.manager.session.account.username
         if proto == PROTO_SSH:
-            return 'ssh://hg@bitbucket.org/' + self.full_name
+            username = 'hg'
+            return 'ssh://{}@bitbucket.org/{}'.format(username, self.full_name)
 
-        return 'https://{}@bitbucket.org/{}'.format(self.manager.session.account.username, self.full_name)
+        return 'https://{}@bitbucket.org/{}'.format(username, self.full_name)
 
     def _do_get_origin_by_cmd(self):
         return self.api_repo.hg_paths()['default']
@@ -386,7 +432,8 @@ class HgRepo(Repo):
     def _do_clone(self, proto, destdir):
         def local_changes():
             changes = self.api_repo.hg_status()
-            del changes['?']
+            if '?' in changes:
+                del changes['?']
             return list(itertools.chain(*changes.values()))
 
         if not os.path.exists(destdir):
@@ -395,7 +442,7 @@ class HgRepo(Repo):
 
             logger.debug("cloning %s", origin)
             hgapi.hg_clone(origin, destdir, *self._custom_args)
-            return True
+            return SyncStatus.CLONED
 
         if not os.path.isdir(destdir) or self != HgRepo.from_dir(self.manager, destdir):
             raise exc.UnrelatedContent(destdir)
@@ -403,17 +450,48 @@ class HgRepo(Repo):
         if local_changes():
             raise exc.UncommitedLocalChanges(destdir)
 
-        self.pull(proto)
+        return self.pull(proto)
 
     def pull(self, proto):
         origin = self._apply_pwd(self._get_origin_by_cmd(proto))
+
+        try:
+            self.api_repo.hg_command(*("incoming -q {}".format(origin).split()))
+        except hgapi.hgapi.HgException:
+            return SyncStatus.UPDATE
+
         self.api_repo.hg_pull(origin)
         self.api_repo.hg_update('tip')
+        return SyncStatus.PULLED
 
     def push(self, proto):
         origin = self._apply_pwd(self._get_origin_by_cmd(proto))
         self.api_repo.hg_push(origin)
 
+    def commited(self):
+        out, err = hg_cmd(self.dirname, 'summary')
+        commit = [line for line in out.split('\n') if line.startswith('commit:')][0]
+        return '(clean)' in commit
+
+    def pushed(self, tries=0):
+        try:
+            self.api_repo.hg_command('outgoing')
+            return 'dirty'
+        except hgapi.hgapi.HgException as e:
+            error = str(e)
+            if 'no changes found' in error:
+                return 'synced'
+
+            if 'repository does not exist' in error:
+                return 'missing'
+
+            if 'Connection reset by peer' in error:
+                return 'cant-connect'
+
+            print('\nwhen push {} ...'.format(self.dirname))
+            print(e)
+            return 'error'
+
 
 class GitRepo(Repo):
     def __init__(self, *args, **kargs):
@@ -458,7 +536,8 @@ class GitRepo(Repo):
     def _do_clone(self, proto, destdir):
         def local_changes():
             changes = self.api_repo.git_status()
-            del changes['??']
+            if '??' in changes:
+                del changes['??']
             return list(itertools.chain(*changes.values()))
 
         if not os.path.exists(destdir):
@@ -467,7 +546,7 @@ class GitRepo(Repo):
 
             logger.debug("cloning %s", origin)
             gitapi.git_clone(origin, destdir)
-            return True
+            return SyncStatus.CLONED
 
         if not os.path.isdir(destdir) or self != GitRepo.from_dir(self.manager, destdir):
             raise exc.UnrelatedContent(destdir)
@@ -475,16 +554,45 @@ class GitRepo(Repo):
         if local_changes():
             raise exc.UncommitedLocalChanges(destdir)
 
-        self.pull(proto)
+        return self.pull(proto)
 
     def pull(self, proto):
         origin = self._apply_pwd(self._get_origin_by_cmd(proto))
+        self.api_repo.git_fetch()
+        if not self.api_repo.git_command(*("log ..origin".split())):
+            return SyncStatus.UPDATE
+
         self.api_repo.git_pull(origin)
+        return SyncStatus.PULLED
 
     def push(self, proto):
         origin = self._apply_pwd(self._get_origin_by_cmd(proto))
         self.api_repo.git_push(origin)
 
+    def commited(self):
+        # print "--- git commited", self.dirname
+        out, err = git_cmd(self.dirname, 'status', '--porcelain')
+        pending = [line for line in out.split('\n')
+                   if line.strip() and not line.startswith('??')]
+        # print pending, bool(pending)
+        return not bool(pending)
+
+    def pushed(self):
+        try:
+            out, err = git_cmd(self.dirname, 'push', '--dry-run')
+            if 'Everything up-to-date' in err:
+                return 'synced'
+            else:
+                return 'dirty'
+        except gitapi.gitapi.GitException as e:
+            if 'Could not read from remote repository' in str(e):
+                return 'cant-connect'
+
+            print('\nwhen push {} ...'.format(self.dirname))
+            print(e)
+
+        return 'error'
+
 
 def check_full_name(full_name):
     msg = "repo name should match 'owner/slug' format"
@@ -496,15 +604,16 @@ def check_full_name(full_name):
 
 
 class RepoManager(object):
-    def __init__(self, account=None):
+    def __init__(self, account=None, num_workers=10):
         self.session = Session(account)
         self._keymanager = None
         self._pool = None
+        self._num_workers = num_workers
 
     @property
     def pool(self):
         if self._pool is None:
-            self._pool = SimpleThreadPool(10)
+            self._pool = SimpleThreadPool(self._num_workers)
 
         return self._pool
 
@@ -532,8 +641,9 @@ class RepoManager(object):
         def get_page_urls(url):
             "generate urls for each page (in paginated reply)"
             base = self.session.http_get(url)
-            npages = 2 + (base['size'] / base['pagelen'])
-            for i in range(1, npages):
+            npages = 1 + (int(math.ceil(base['size'] // base['pagelen'])))
+
+            for i in range(1, npages + 1):
                 yield "{}?page={}".format(url, i)
 
         def threaded_page_request(url):
@@ -621,10 +731,10 @@ class SshKeyManager(object):
     def add_key(self, label, keyfile):
         try:
             retval = self.session.http_post(
-                bb_api1_uri('ssh-keys'.format(self.session.account.username)),
+                bb_api1_uri('ssh-keys'),
                 data=dict(label=label, key=keyfile.read()))
         except exc.RemoteException as e:
-            if e.reason == 'BAD_REQUEST':
+            if e.reason.upper() == 'BAD_REQUEST':
                 raise exc.NoValidKey()
             raise
 
