@@ -6,8 +6,10 @@ import getpass
 import logging
 import webbrowser
 import argparse
-import OpenSSL
-from functools import partial
+import collections
+import datetime
+import dateutil.parser
+# import OpenSSL
 
 logging.basicConfig()
 logger = logging.getLogger('bucket')
@@ -31,6 +33,40 @@ def msg(text, fd=None):
     fd = fd or args.stdout
     fd.write(str(text) + '\n')
 
+# import math
+# http://stackoverflow.com/questions/5194057/better-way-to-convert-file-sizes-in-python
+# def convertSize(size):
+#     size_name = ("KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB")
+#     i = int(math.floor(math.log(size, 1024)))
+#     p = math.pow(1024, i)
+#     s = round(size / p, 2)
+#     if (s > 0):
+#         return '%s %s' % (s, size_name[i])
+#     else:
+#         return '0B'
+
+
+def convert_size(bytes):
+    k = float(bytes) / 1000
+    return "{:.0f} KB".format(k)
+
+
+def user_confirm(text, valid_answers):
+    try:
+        answer = raw_input(text)
+    except KeyboardInterrupt:
+        canceled()
+
+    if answer not in valid_answers:
+        canceled()
+
+    return answer
+
+
+def confirm_irrecoverable_operation():
+    user_confirm("This is an IRRECOVERABLE operation!!\nAre you sure? (write uppercase 'yes'): ",
+                 valid_answers=['YES'])
+
 
 class readable_dir(argparse.Action):
     def __call__(self, parser, namespace, values, option_string=None):
@@ -66,25 +102,31 @@ def repo2full_name(repo):
 
 
 def get_repo():
-    full_name = repo2full_name(args.repo)
+    repo = args.repo
+
+    if repo and repo.count('/') == 0:
+        repo = '/' + repo
+
+    full_name = repo2full_name(repo)
 
     try:
         return api.Repo.from_remote(args.manager, full_name)
 
     except exc.BasicException:
-        if not args.repo.startswith('/'):
+        if not repo.startswith('/'):
             raise
 
-        return get_alternate_repo(args.repo)
+        return get_alternate_repo(repo)
 
 
 def get_alternate_repo(repo):
-    full_name = repo2full_name(args.repo)
-    owner = full_name.split('/')[0]
+    full_name = repo2full_name(repo)
+    owner, name = full_name.split('/')
 
-    for owner in [owner] + get_config_owners(owner):
+    owners = [owner] + get_config_owners(owner)
+    for owner in owners:
         try:
-            return api.Repo.from_remote(args.manager, full_name)
+            return api.Repo.from_remote(args.manager, owner + '/' + name)
         except exc.BasicException:
             logger.info("trying {}: FAIL".format(full_name))
 
@@ -94,12 +136,12 @@ def get_alternate_repo(repo):
 def get_config_owners(owner):
     try:
         accounts = args.ui.account.values()
-    except KeyError:
+    except AttributeError:
         return []
 
     retval = []
     for account in accounts:
-        if account.count(':') > 1:
+        if account.count(':') > 2:
             logger.warning("wrong account format: '%s'", account)
 
         retval.append(account.split(':')[0])
@@ -111,11 +153,38 @@ def get_config_owners(owner):
 
 
 def repo_info():
+    def format_delta(strtimestamp):
+        day = 24 * 60 * 60
+        week = 7 * day
+
+        updated_on = dateutil.parser.parse(strtimestamp)
+        updated_on = updated_on.replace(tzinfo=None)
+        now = datetime.datetime.now()
+        delta = now - updated_on
+
+        strtimestamp = updated_on.strftime('%Y-%m-%d %H:%M')
+        if delta.total_seconds() > week:
+            return strtimestamp
+
+        reldelta = dateutil.relativedelta.relativedelta(now, updated_on)
+        if delta.total_seconds() > day:
+            strdelta = "{} days ago".format(reldelta.days)
+        else:
+            strdelta = "{hours:02}h{minutes:02} ago".format(**reldelta.__dict__)
+
+        return "{} ({})".format(strtimestamp, strdelta)
+
     repo = get_repo()
+    info = collections.OrderedDict()
 
     for atr in 'full_name name description scm is_private size updated_on'.split():
-        key = atr.replace('_', ' ') + ':'
         value = str(repo[atr])
+        info[atr] = value
+
+    info['updated_on'] = format_delta(info['updated_on'])
+
+    for key, value in info.items():
+        key = key.replace('_', ' ') + ':'
         if not value.strip():
             continue
 
@@ -125,28 +194,24 @@ def repo_info():
 def repo_create():
     full_name = repo2full_name(args.repo)
 
-    scm = 'git' if args.create_git else 'hg'
+    scm = 'git' # if args.create_git else 'hg'
 
     if not args.create_yes:
-        answer = raw_input("create '{}' ({}) repository? (Y/n): ".format(full_name, scm))
-        if answer.lower() not in ['', 'y', 'yes']:
-            canceled()
+        user_confirm("create '{}' ({}) repository? (Y/n): ".format(full_name, scm),
+                     valid_answers=['', 'y', 'yes', 'Y', 'YES'])
 
     else:
         msg("creating '{}' ({})".format(full_name, scm))
 
     args.manager.repo_create(full_name, scm=scm)
 
-    if args.create_local:
-        msg("pushing local content to '{}'".format(full_name))
-        repo = api.Repo.from_dir()
-        repo.push(args.manager.repo_info(full_name), args.create_proto, args.account)
+    # if args.create_local:
+    #     msg("pushing local content to '{}'".format(full_name))
+    #     repo = api.Repo.from_dir(args.manager)
+    #     repo.push(args.create_proto)
 
-
-def confirm_irrecoverable_operation():
-    answer = raw_input("This is an IRRECOVERABLE operation!!\nAre you sure? (write uppercase 'yes'): ")
-    if answer != 'YES':
-        canceled()
+    if args.create_clone:
+        repo_clone()
 
 
 def repo_delete():
@@ -185,24 +250,33 @@ def repo_ls():
     if args.ls_scm:
         item = u'{scm:3} - ' + item
 
+    total_size = 0
     if args.ls_size:
         item = u'{size:>13} - ' + item
 
     lines = []
     for repo in args.manager.names2repos(names):
-        line = item.format(**repo.as_dict()) + u'\n'
+        repodata = repo.as_dict()
+        repodata['size'] = convert_size(repodata['size'])
+        line = item.format(**repodata) + u'\n'
         lines.append(line)
 
+        if args.ls_size:
+            total_size += repo['size']
+
     for line in lines:
         args.stdout.write(line.encode('utf-8'))
 
+    if args.ls_size:
+        msg("\n-- total storage: {}".format(convert_size(total_size)))
+
 
 def repo_clone():
     repo = get_repo()
 
     destdir = os.path.join(args.clone_destdir or os.path.join(os.getcwd()), repo.slug)
 
-    msg("Clonning '{}' to '{}' ({}/{}) ".format(
+    msg("clonning '{}' to '{}' ({}/{}) ".format(
         repo.full_name, destdir, repo.scm, args.clone_proto))
 
     if not repo.clone(args.clone_proto, destdir):
@@ -218,7 +292,7 @@ def relpath(name):
 
 
 def put_point():
-    args.stderr.write('.')
+    args.stderr.write('-')
     args.stderr.flush()
 
 
@@ -227,15 +301,22 @@ def repo_sync():
         destdir = os.path.join(args.clone_destdir or os.path.join(os.getcwd()), repo.slug)
         proto_info = "({}/{})".format(repo.scm, args.clone_proto)
 
-        msg("Syncing {0:11} {1}".format(
-            proto_info, repo.full_name, relpath(destdir)))
-
         try:
-            repo.clone(args.clone_proto, destdir)
-        except (exc.CloneError, exc.UnrelatedContent, OpenSSL.SSL.ZeroReturnError) as e:
+            result = repo.clone(args.clone_proto, destdir)
+            if result != api.SyncStatus.UPDATE or args.verbosity:
+                msg("Syncing {0:11} ({3}) {1}".format(
+                    proto_info, repo.full_name, relpath(destdir), result))
+
+        except (exc.CloneError, exc.UnrelatedContent) as e:
             logger.error(e)
 
-    # FIXME: repeated in repo_ls
+        # , OpenSSL.SSL.ZeroReturnError
+
+    def is_valid_destdir(path):
+        return path and \
+            os.path.isdir(path)
+
+    # FIXME: same as repo_ls
     owner = args.sync_owner
     if owner is None:
         if args.account.is_anonymous():
@@ -245,6 +326,7 @@ def repo_sync():
 
     nrepos = args.manager.get_repo_amount(owner)
     msg("-- owner '{}' has {} repositories".format(owner, nrepos))
+    assert is_valid_destdir(args.clone_destdir), "Invalid destination directory: {}".format(args.clone_destdir)
     msg("-- target directory is: '{}'".format(os.path.abspath(args.clone_destdir)))
 
     repos = args.manager.repo_list(owner, step=put_point)
@@ -256,8 +338,7 @@ def repo_sync():
 
 
 def repo_website():
-    repo = args.manager.repo_info(repo2full_name(args.repo))
-    url = repo.webpage
+    url = get_repo().webpage
 
     if args.site_settings:
         url += '/admin'
@@ -271,35 +352,80 @@ def repo_status():
     not_bb_repos = []
     not_repos = []
 
-    for d in os.listdir(args.clone_destdir):
-        put_point()
+    def get_repo_paths():
+        retval = []
+        for d in sorted(os.listdir(args.clone_destdir)):
+            path = os.path.join(args.clone_destdir, d)
+            if not os.path.isdir(path):
+                continue
 
-        path = os.path.join(args.clone_destdir, d)
-        if not os.path.isdir(path):
-            continue
+            retval.append(path)
 
-        try:
-            local_repo = api.Repo.from_dir(path)
-            if local_repo.path:
-                bb_repos.append(local_repo)
-            else:
-                not_bb_repos.append(local_repo)
+        return retval
 
-        except exc.NoSuchRepo:
-            not_repos.append(d)
+    def map_get_repos(paths):
+        def get_repo(path):
+            put_point()
+            try:
+                local_repo = api.Repo.from_dir(args.manager, path)
+                if local_repo.full_name:
+                    bb_repos.append(local_repo)
+                else:
+                    not_bb_repos.append(local_repo)
+
+            except exc.NoSuchRepo:
+                not_repos.append(os.path.split(path)[1])
+
+        pool = SimpleThreadPool(40)
+        retval = pool.map(get_repo, paths)
+        print(' ok')
+        return retval
+
+    def get_statuses(repos):
+        def get_sync_status(repo):
+            put_point()
+            commit_states = {False: '✘', True: '✔', None: '?'}
+            push_states = {
+                'dirty':        '✘',
+                'synced':       '✔',
+                'cant-connect': 'r',
+                'error':        'e',
+                'missing':      'm',
+                'unrelevant':   '-',
+            }
+            commited = repo.commited()
+            if commited is False:
+                return commit_states[False], push_states['unrelevant']
+
+            return [commit_states[commited], push_states[repo.pushed()]]
+
+        pool = SimpleThreadPool(2)
+        retval = pool.map(get_sync_status, repos)
+        print(' ok')
+        return retval
+
+    print(args.clone_destdir)
+    paths = get_repo_paths()
+    map_get_repos(paths)
+    bb_repos.sort()
+    repo_statuses = get_statuses(bb_repos)
 
     print
+    msg(" +-commited")
+    msg(" |,-pushed=remote-issue(r)|missing-repo(m)|unrelevant(-)|other-error(e)")
+    for r, status in zip(bb_repos, repo_statuses):
+        msg(" {0[0]}{0[1]} {1:<3} - {2:<30} - {3}".format(
+            status, r.scm, r.basename, r.full_name))
 
-    for r in bb_repos:
-        msg("+ {:<3} - {:<30} - {}".format(r.scm, os.path.split(r.dirname)[1], r.path))
-
-    msg("\n -- Not bitbucket.org repos:")
+    msg("\n -- Not bitbucket.org repos --")
     for r in not_bb_repos:
-        msg("- {:<3} - {:<30} - {}".format(r.scm, os.path.split(r.dirname)[1], r.origin))
+        msg(" - {:<3} - {:<30} - {}".format(r.scm, r.basename, r.origin))
 
-    msg('\n -- Not a repository:')
+    msg('\n -- Not a repository --')
     for r in not_repos:
-        msg(r)
+        msg(" - {}".format(r))
+
+    print
 
 
 def ssh_ls():
@@ -372,16 +498,16 @@ def setup_parser_create(commands):
     subparser.add_argument('repo')
     subparser.add_argument('--yes', '-y', dest='create_yes', action='store_true',
                            help="don't ask for confirmation")
-    subparser.add_argument('--git', dest='create_git', action='store_true',
-                           help='create a git repository')
-    subparser.add_argument('--from-local', dest='create_local', action='store_true',
-                           help="push current content to the created repo")
+    # subparser.add_argument('--from-local', dest='create_local', action='store_true',
+    #                        help="push current content to the created repo")
+
+    subparser.add_argument('--clone', dest='create_clone', action='store_true',
+                           help="create repo, then clone (proto and destdir depends on config)")
 
     # --hg
     # --public
     # --private
     # --desc
-    # --clone
 
 
 def setup_parser_delete(commands):
@@ -401,8 +527,7 @@ def add_proto_args(parser, dest):
 
 def add_destdir_arg(parser):
     parser.add_argument(
-        '--destdir', '-d', dest='clone_destdir', metavar='dirname',
-        action=readable_dir,
+        '--destdir', '-d', dest='clone_destdir', metavar='dirname', action=readable_dir,
         help='the directory where your repositories are')
 
 
@@ -481,6 +606,8 @@ def parse_args(argv=None, ns=None, stdout=None):
     setup_verbosity()
 
     args.clone_proto = args.get('clone_proto') or api.PROTO_HTTPS
+    args.clone_destdir = args.get('clone_destdir') or "."
+    args.clone_destdir = os.path.expanduser(args.clone_destdir)
 
     args.account = setup_account(args.account)
     logger.info("Authenticated as: %s", args.account.username)
@@ -496,7 +623,9 @@ def load_config_file(argv):
     auxp.add_argument('--config', '-c', default=os.path.expanduser('~/.bucket'))
     values, pos = auxp.parse_known_args(argv)
 
-    if not os.path.isfile(values.config):
+    if values.config == '/dev/null':
+        logging.warning("no config set")
+    elif not os.path.isfile(values.config):
         logger.error("bad file: %s", values.config)
     else:
         parser.load_config_file(values.config)
@@ -533,7 +662,8 @@ def setup_account(account):
         username = account
         password = get_password(account)
 
-    return api.Account(username, password)
+    retval = api.Account(username, password)
+    return retval
 
 
 setup_parser()
