@@ -5,9 +5,20 @@
 import urlparse
 
 
+def parse(url):
+    retval = urlparse.urlparse(url)
+    if not retval.scheme and url.startswith('git@'):
+        url = url.replace(':', '/')
+        url = 'git://' + url
+        # print("--", url)
+        return urlparse.urlparse(url)
+
+    return retval
+
+
 def change_login(url, user='', password=''):
     """
-    >>> change_user(urlparse.urlparse("http://guido@www.cwi.nl:80/python.html")
+    >>> change_login(urlparse.urlparse("http://guido@www.cwi.nl:80/python.html")
     >>> _.geturl()
     'http://www.cwi.nl:80/python.html'
     """
